/*
    This file is part of the source of talk64 V1.5,
    a PC - DoctorV64 communications program
    Copyright (C) 1997  Andreas Sterbenz (stan@sbox.tu-graz.ac.at)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "global.h"
#include "interact.h"
#include "v64proto.h"
#include "parsys.h"

#define MAX_ARGS 10

#define DUMPMODE_HEX 0
#define DUMPMODE_ASC 1

char *hdumpLine(unsigned char *buffer, long offset, long len)
{
  static char line[80];       /* static ! */
  char *c;
  int i;

  c = line;
  c += sprintf(c, "%08lX:  ", offset);
  for( i=0; i<16; i++ ) {
    if( i==8 ) c += sprintf(c, " ");
    if( i<len ) {
      c += sprintf(c, "%02X ", buffer[i]);
    } else {
      c += sprintf(c, "   ");
    }
  }
  c += sprintf(c, " ");
  for( i=0; i<len; i++ ) {
    c += sprintf(c, "%c", printableChar(buffer[i]));
  }
  return line;
}

char *adumpLine(unsigned char *buffer, long offset, long len)
{
  static char line[80];       /* static ! */
  char *c;
  int i;

  c = line;
  c += sprintf(c, "%08lX:  ", offset);
  for( i=0; i<len; i++ ) {
    c += sprintf(c, "%c", printableChar(buffer[i]));
  }
  return line;
}

typedef struct {
  int n;
  char rawcommand[MAX_STRING_LENGTH];
  char p[MAX_ARGS][MAX_STRING_LENGTH];
} parsedCommand;

parsedCommand simpleCommandParser(char command[])
{
  char *c=command;
  parsedCommand p;
  int i, j;

  strcpy(p.rawcommand, command);
  p.n = 0;
  do {
    while( isspace(*c) ) c++;
    if( *c == ',' ) c++;
    while( isspace(*c) ) c++;
    if( *c != '\0' ) {
      i = p.n++;
      j = 0;
      while( (*c != '\0') && (!isspace(*c)) && (*c != ',') ) {
        if( j < MAX_STRING_LENGTH-1 ) p.p[i][j++] = *c;
        c++;
      }
      p.p[i][j] = '\0';
    }
  } while( *c != '\0' );

  return p;
}

void printParsedCommand(parsedCommand p)
{
  int i;

  printf("Number: %d\n", p.n);
  for( i=0; i<p.n; i++ ) {
    printf("%d: '%s'\n", i, p.p[i]);
  }
}

int v64Print(unsigned int baseport, parsedCommand p)
{
  unsigned char buffer2[16];
  unsigned char line, column;
  int i;

  if( p.n < 4 ) {
    printf("Error in use of 'print', see help.\n");
    return 0;
  }
  line = atoi(p.p[1]);
  column = atoi(p.p[2]);
  sendHeaderExtPrint(baseport, line, column);
  buffer2[0] = ' ';
  for( i=3; i<p.n; i++ ) {
    if( sendBytes(baseport, strlen(p.p[i]), p.p[i], 0) != 0 ) return 1;
    if( sendBytes(baseport, 1, (char *)buffer2, 0) != 0 ) return 1;
  }
  buffer2[0] = 0xff;
  if( sendBytes(baseport, 1, (char *)buffer2, 0) != 0 ) return 1;
  return 0;
}

int v64Call(unsigned int baseport, parsedCommand p)
{
  int address;
  unsigned char buffer2[15];

  if( p.n != 2 ) {
    printf("Error in use of 'call', see help.\n");
    return 0;
  }
  if( sscanf(p.p[1], "%x%s", &address, buffer2) != 1 ) return 1;
  return sendHeaderExtCall(baseport, address);
}

long prevOffset=0, prevLen=20*16;
int prevSwapped=0;

int v64Dump(unsigned int baseport, parsedCommand p, int dumpmode)
{
  long offset, len;
  int swap;

  if( p.n > 3 ) {
    printf("Error in use of '%cdump', see help.\n", (dumpmode == DUMPMODE_HEX) ? 'h' : 'a');
    return 0;
  }
  if( (p.n > 1) && (p.p[1][0] != '\0') ) {
    if( sscanf(p.p[1], "%lx%s", &offset, buffer1) != 1 ) return 1;
  } else {
    offset = prevOffset;
  }
  if( (p.n > 2) && (p.p[2][0] != '\0')) {
    if( sscanf(p.p[2], "%lx%s", &len, buffer1) != 1 ) return 1;
  } else {
    len = prevLen;
  }
  if( ((offset|len) & 1) != 0 ) {
    printf("Address and length have to be even!\n");
    return 1;
  }
  prevOffset = offset + len;
  prevLen = len;
  if( sendHeaderExtRec(baseport, offset, len) != 0 ) return 1;
  swap = (offset < 0x80000000) ? 1 : 0;
  for( ;; ) {
    unsigned int k;
    static unsigned int n=0;
    unsigned char *b;

    n = min2(len, BUFSIZE_INT);
    if( n <= 0 ) break;
    if( receiveBytes(baseport, n, buffer1, swap) != 0 ) return 1;
    b = (unsigned char *)buffer1;
    for( ;; ) {
      if( dumpmode == DUMPMODE_HEX) {
        k = min2(n, 16);
        printf("%s\n", hdumpLine(b, offset, k));
      } else { /* dumpmode == DUMPMODE_ASC */
        k = min2(n, 64);
        printf("%s\n", adumpLine(b, offset, k));
      }
      offset += k;
      n -= k;
      len -= k;
      b += k;
      if( n <= 0 ) break;
    }
  }
  return 0;
}

int v64Help(void)
{
  printf("Talk64 interactive mode - short command summary.\n\n");
  printf("help                                  print this help text\n");
  printf("quit                                  exit Talk V64\n");
  printf("hdump address len                     print a hexdump of the V64 DRAM memory\n");
  printf("adump address len                     print a ascii dump of the V64 DRAM memory\n");
  printf("upload filename address len byteorder upload len bytes from filename to address\n");
  printf("download filename address len b-order download len bytes to filename\n");
  printf("call address                          call V64 BIOS subroutine at address\n");
  printf("print line column some text           print some text on the V64 screen\n");
  return 0;
}

int v64Upload(unsigned int baseport, parsedCommand p)
{
  FILE *file2;
  unsigned char buffer2[15];
  unsigned char buffer3[12];
  long offset, len, flen2;
  int swapped=-1, v64format;
  int detect=1;

  if( (p.n < 2) || (p.n > 5) ) {
    printf("Error in use of 'upload', see help.\n");
    return 0;
  }
  if( (file2 = fopen(p.p[1], "rb")) == NULL ) {
    printf("Could not open file '%s' for reading.\n", p.p[1]);
    return 0;
  }
  if( fread(buffer1, 1, 12, file2) != 12 ) detect = 0;
  fseek(file2, 0, SEEK_END);
  flen2 = ftell(file2);
  fseek(file2, 0, SEEK_SET);
  if( (p.n > 2) && (p.p[2] != '\0') ) {
    if( sscanf(p.p[2], "%lx%s", &offset, buffer2) != 1 ) return 1;
    LONG2BYTES(offset, &buffer2[0], 0);
  } else {
    offset = 0;
  }
  if( (offset & 1) != 0 ) {
    printf("Address has to be even!\n");
    return 1;
  }
  if( offset != 0 ) detect = 0;
  if( (p.n > 3) && (p.p[3] != '\0') ) {
    if( sscanf(p.p[3], "%lx%s", &len, buffer2) != 1 ) return 1;
    LONG2BYTES(len, &buffer2[4], 0);
  } else {
    len = flen2;
  }
  if( (len & 1) != 0 ) {
    printf("WARNING: Given length is odd, last byte will be dropped!\n");
    len--;
  }
  if( (p.n > 4) ) {
    swapped = atoi(p.p[4]);
  } else {
    if( detect != 0 ) {
      swapped = guessByteOrder(buffer3);
    }
    if( swapped == -1 ) {
      swapped = prevSwapped;
    }
  }
  swapped &= 1;
  prevSwapped = swapped;
  printf("Uploading 0x%lX bytes from file '%s' to memory address 0x%lX...\n", len, p.p[1], offset);
  printf("Assuming %s byteorder.\n", (swapped==0) ? "original (not swapped)" : "V64 (swapped)");
  v64format = swapped ^ 1;
  if( sendHeaderExtSend(baseport, offset, len) != 0 ) return 1;
  sendBytesFromFile(baseport, file2, len, v64format);
  fclose(file2);
  return 0;
}

int v64Download(unsigned int baseport, parsedCommand p)
{
  FILE *file2;
  unsigned char buffer2[15];
  long offset, len;
  int swapped=-1, v64format;

  if( (p.n < 4) || (p.n > 5) ) {
    printf("Error in use of 'download', see help.\n");
    return 0;
  }
  if( (p.n > 2) && (p.p[2] != '\0') ) {
    if( sscanf(p.p[2], "%lx%s", &offset, buffer2) != 1 ) return 1;
    LONG2BYTES(offset, &buffer2[0], 0);
  } else {
    return 1;
  }
  if( (p.n > 3) && (p.p[3] != '\0') ) {
    if( sscanf(p.p[3], "%lx%s", &len, buffer2) != 1 ) return 1;
    LONG2BYTES(len, &buffer2[4], 0);
  } else {
    return 1;
  }
  if( ((offset|len) & 1) != 0 ) {
    printf("Address and length have to be even!\n");
    return 1;
  }
  if( (p.n > 4) ) {
    swapped = atoi(p.p[4]);
  } else {
    swapped = prevSwapped;
  }
  swapped &= 1;
  prevSwapped = swapped;
  if( (file2 = fopen(p.p[1], "wb")) == NULL ) {
    printf("Could not open file '%s' for writing.\n", p.p[1]);
    return 0;
  }
  printf("Downloading 0x%lX bytes starting at memory address 0x%lX to file '%s'...\n", len, offset, p.p[1]);
  printf("Assuming %s byteorder.\n", (swapped==0) ? "original (not swapped)" : "V64 (swapped)");
  v64format = swapped ^ 1;
  if( sendHeaderExtRec(baseport, offset, len) != 0 ) return 1;
  receiveBytesToFile(baseport, file2, len, v64format);
  fclose(file2);
  return 0;
}

int v64Edit(unsigned int baseport, parsedCommand p)
{
  unsigned char buffer2[15];
  long offset;
  long len;
  int swap;

  if( p.n != 2 ) {
    printf("Error in use of 'edit', see help.\n");
    return 0;
  }
  if( sscanf(p.p[1], "%lx%s", &offset, buffer2) != 1 ) return 1;
  len = BUFSIZE_INT;
  if( ((offset|len) & 1) != 0 ) {
    printf("Address has to be even!\n");
    return 1;
  }
  LONG2BYTES(offset, &buffer2[0], 0);
  LONG2BYTES(len, &buffer2[4], 0);
  swap = (offset < 0x80000000) ? 1 : 0;
  if( sendHeaderExtRec(baseport, offset, len) != 0 ) return 1;
  if( receiveBytes(baseport, len, buffer1, swap) != 0 ) return 1;

  setInputMode(INPUT_MODE_NOECHO);
  {
    int i, j, key1;
    int edithex = 1;
    unsigned char *b;
    char *c;
    int highnibble = 1;
    long currentofs;

    b = 0;
    i = 0;
    currentofs = 0;
    for( ;; ) {
      while( i >= 16 ) {
        printf("\r%s\n", hdumpLine(b, currentofs, 16));
        i -= 16;
        currentofs += 16;
        b += 16;
      }
      if( i < 0 ) {
        printf("\n");
        while( i < 0 ) {
          i += 16;
          currentofs -= 16;
          b -= 16;
        }
      }
      if( b < (unsigned char *)buffer1 ) {
        b = (unsigned char *)buffer1;
        i = 0;
        currentofs = offset;
      }
      if( currentofs >= offset + len ) break;
      c = hdumpLine(b, currentofs, 16);
      if( edithex ) {
        j = 8 + 3 + 3*i + (i>7 ? 1 : 0) + (highnibble ^ 1);
      } else {
        j = 8 + 3 + 3*16 + 1 + 1 + i;
      }
      printf("\r%s\r%.*s", c, j, c);
      fflush(stdout);
      key1 = getKbChar();
      if( key1 == '\x1b' ) {
        break;
      } else if( (key1 == '\n') || (key1 == '\r') ) {
        i=16;
        highnibble = 1;
        continue;
      } else if( key1 == '\t' ) {
        edithex ^= 1;
        highnibble = 1;
        continue;
      } else if( key1 == '\b' ) {
        if( edithex ) {
          highnibble ^= 1;
          if( highnibble == 0 ) i--;
        } else {
          i--;
        }
        continue;
      } else if( edithex ) {
        if( key1 == ' ' ) {
          i++;
          highnibble = 1;
          continue;
        } else if( isxdigit(key1) ) {
          int nibble = 0;
          if( isdigit(key1) ) {
            nibble = key1 - '0';
          } else {
            nibble = 10 + tolower(key1) - 'a';
          }
          if( highnibble ) {
            b[i] = (nibble << 4) | (b[i] & 0x0f);
          } else {
            b[i] = (b[i] & 0xf0) | nibble;
          }
          highnibble ^= 1;
          if( highnibble == 1 ) i++;
        }
        continue;
      } else { /* editascii */
        b[i++] = key1;
      }
    }
  }
  printf("\nWrite back changes? (y/n)");
  fflush(stdout);
  for( ;; ) {
    int c = tolower(getKbChar());
    setInputMode(INPUT_MODE_ECHO);
    if( c == 'n' ) {
      printf(" no, changes cancelled.\n");
      break;
    } else if( c == 'y' ) {
      printf(" yes, saving changes.\n");
      if( sendHeaderExtSend(baseport, offset, len) != 0 ) return 1;
      if( sendBytes(baseport, len, buffer1, swap) != 0 ) return 1;
      break;
    }
    setInputMode(INPUT_MODE_NOECHO);
  }
  return 0;
}

int interactiveMode(unsigned int port)
{
  char userin[MAX_STRING_LENGTH];
  parsedCommand p;

  if( biosVersion < 0x0130 ) {
    fprintf(stderr, "%s: Interactive mode not available with BIOS versions prior to 1.30.\n", progname);
    exit(1);
  }
  initCommunication(port);
  if( (sendHeaderExtRec(port, 0, -16) != 0) ||
      (receiveBytes(port, 16, userin, 0) != 0) ) {
    fprintf(stderr, "%s: Some communications problem occured.\n", progname);
    fprintf(stderr, "Chances are that you have some pre 1.30 BIOS that does no support the features required for interactive mode.\n");
    exit(1);
  }
  p.n = -1;
  printf("Enter command, 'help' for help, 'quit' to end.\n");
  for( ;; ) {
    char *c;

    printf(">");
    fflush(stdout);
    if( fgets(userin, MAX_STRING_LENGTH, stdin) == 0 ) {
      c = "quit";
    } else {
      c = userin;
    }
    if( (strcmp(c, " \n") != 0) || (p.n < 0) ) {
      p = simpleCommandParser(c);
    } else {
      /* p = p -- leave previous command in place */
    }
    if( p.n == 0 ) continue;
    if( (strcmp(p.p[0], "quit") == 0) ||
        (strcmp(p.p[0], "exit") == 0) ||
        (strcmp(p.p[0], "bye") == 0) ) {
      break;
    } else if( strcmp(p.p[0], "help") == 0 ) {
      v64Help();
    } else if( strcmp(p.p[0], "print") == 0 ) {
      v64Print(port, p);
    } else if( strcmp(p.p[0], "call") == 0 ) {
      v64Call(port, p);
    } else if( strcmp(p.p[0], "hdump") == 0 ) {
      v64Dump(port, p, DUMPMODE_HEX);
    } else if( strcmp(p.p[0], "adump") == 0 ) {
      v64Dump(port, p, DUMPMODE_ASC);
    } else if( strcmp(p.p[0], "edit") == 0 ) {
      v64Edit(port, p);
    } else if( strcmp(p.p[0], "upload") == 0 ) {
      v64Upload(port, p);
    } else if( strcmp(p.p[0], "download") == 0 ) {
      v64Download(port, p);
    } else {
      printf("Unknown command, enter 'help' for help, 'quit' to quit interactive mode.\n");
    }
  }
  return 0;
}
