/*
    This file is part of the source of talk64 V1.5,
    a PC - DoctorV64 communications program
    Copyright (C) 1997  Andreas Sterbenz (stan@sbox.tu-graz.ac.at)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "global.h"
#include "interact.h"
#include "v64proto.h"
#include "parsys.h"     /* definition of 'system-dependent' functions */

  char buffer1[BUFSIZE];
  char *progname;
  int biosVersion;

void usage(char *name)
{
  fflush(stdout);
  fprintf(stderr, "Usage: %s [-o|-s] [-pPort] [-lLength] [-send|-rec|-int] <File> [-a Arguments]\n\n", name);
  fprintf(stderr, "A PC-to-V64 upload/ download/ general communications program.\n");
  exit(2);
}

int guessByteOrder(unsigned char rombytes[12])
{
  if( BYTES2LONG(rombytes, 0) == HEADER_MAGIC ) {
    return 0;
  } else if( BYTES2LONG(rombytes, 1) == HEADER_MAGIC ) {
    return 1;
  }
  if( (rombytes[8] == 0x80) && (rombytes[9] != 0x80) ) {
    return 2;
  } else {
    if( rombytes[9] == 0x80 ) {
      return 3;
    }
  }
  return -1;
}

int sendBytesFromFile(int baseport, FILE *file1, long flen1, int v64format)
{
  long bytessent=0;
  int sendmem=0;

  printf("%8ld bytes (%ld.%02ld Mbit) will be transmitted.\n", flen1, flen1>>17, ((flen1*100)>>17)%100);
  fflush(stdout);
  for( ;; ) {
    unsigned int n;

    if( sendmem == 0 ) {
      n = fread(buffer1, 1, min2(flen1, BUFSIZE), file1);
    }
    if( (sendmem != 0) || (n == 0) ) {
      if( sendmem == 0 ) {
        sendmem = 1;
        memset(buffer1, 0, BUFSIZE);
      }
      n = min2(flen1, BUFSIZE);
    }
    if( n == 0 ) break;
    if( sendBytes(baseport, n, buffer1, v64format) != 0 ) {
      fflush(stdout);
      fprintf(stderr, "\n%s: Timeout while transmitting data to the Doctor V64!\n", progname);
      fprintf(stderr, "Please check cables and connection.\n");
      exit(1);
    }
    bytessent += n;
    printf("\r%8ld Bytes sent.", bytessent);
    fflush(stdout);
    flen1 -= n;
  }
  printf("\n");
  return 0;
}

int receiveBytesToFile(int baseport, FILE *file1, long flen1, int v64format)
{
  printf("%8ld bytes (%ld.%02ld Mbit) will be transmitted.\n", flen1, flen1>>17, ((flen1*100)>>17)%100);
  fflush(stdout);
  for( ;; ) {
    unsigned int n;
    static long bytesreceived=0;

    n = min2(flen1, BUFSIZE);
    if( n <= 0 ) break;
    if( receiveBytes(baseport, n, buffer1, v64format) != 0 ) {
      fflush(stdout);
      fprintf(stderr, "\n%s: Timeout while receiving data from the Doctor V64!\n", progname);
      fprintf(stderr, "Please check cables and connection.\n");
      exit(1);
    }
    bytesreceived += n;
    printf("\r%8ld Bytes received.", bytesreceived);
    fflush(stdout);
    if( fwrite(buffer1, 1, n, file1) != n ) {
      fprintf(stderr, "\n%s: Error writing to output file!\n", progname);
      exit(1);
    }
    flen1 -= n;
  }
  printf("\n");
  return 0;
}

int sendFile(unsigned int port, char *fname1, long slen, int swapped)
{
  FILE *file1;
  long flen1;
  int v64format;

  fflush(stdout);
  if( (file1 = fopen(fname1, "rb")) == NULL ) {
    fprintf(stderr, "%s: Could not open file '%s' for reading.\n", progname, fname1);
    exit(1);
  }
  if( fread(buffer1, 1, 12, file1) != 12 ) {
    fprintf(stderr, "%s: File too short to be transmitted!\n", progname);
    exit(1);
  }
  fseek(file1, 0, SEEK_END);
  flen1 = ftell(file1);
  fseek(file1, 0, SEEK_SET);
  if( swapped == -1 ) {
    swapped = guessByteOrder((unsigned char *)buffer1);
  }
  if( (swapped == 0) || (swapped == 1) ) {
    printf("The image '%s' is in %s format.\n", fname1, (swapped==0) ? "original (not swapped)" : "V64 (byte-swapped)");
  } else {
    if( (flen1 == 262144) && (slen == -1) ) {
      printf("The file '%s' appears to be a V64 BIOS file, OK.\n", fname1);
      swapped = 1;
    } else {
      if( swapped != -1 ) {
        swapped &= 1;
        printf("WARNING: I am not 100%% certain, but the image '%s' appears to be in %s format.\n", fname1, (swapped==0) ? "original (not swapped)" : "V64 (byte-swapped)");
      } else {
        fflush(stdout);
        fprintf(stderr, "%s: Could not detect type of image '%s', use commandline to override.\n", progname, fname1);
        exit(1);
      }
    }
  }
  if( slen != -1 ) {
    if( slen == -2 ) {
      if( biosVersion < 0x0130 ) {
        slen = (((flen1-1) >>23) +1 ) <<23;
      } else if( biosVersion < 0x0133 ) {
        slen = ( max2( ((flen1-1) >>18) +1, 5) ) <<18;
      } else {
        slen = (((flen1-1) >>18) +1 ) <<18;
      }
    }
  } else {
    slen = flen1;
  }
  v64format = swapped ^ 1;
  initCommunication(port);
  if( sendUploadHeader(port, fname1, slen) != 0 ) {
    fflush(stdout);
    fprintf(stderr, "%s: Error sending transfer headers to the Doctor V64!\n", progname);
    fprintf(stderr, "Please check cables and connection.\n");
    exit(1);
  }
  sendBytesFromFile(port, file1, slen, v64format);
  fclose(file1);
  printf("\nFile transmitted successfully!\n");
  return 0;
}

int receiveFile(unsigned int port, char *fname1, long slen, int swapped)
{
  FILE *file1;
  int v64format;

  if( (file1 = fopen(fname1, "wb")) == NULL ) {
    fflush(stdout);
    fprintf(stderr, "%s: Could not open file '%s' for writing.\n", progname, fname1);
    exit(1);
  }
  if( swapped == -1 ) swapped = 1;
  printf("The image '%s' will be saved in %s format.\n", fname1, (swapped==0) ? "original (not swapped)" : "V64 (byte-swapped)");
  v64format = swapped^1;
  initCommunication(port);
  if( sendDownloadHeader(port, fname1, &slen) != 0 ) {
    fflush(stdout);
    fprintf(stderr, "%s: Error sending transfer headers to the Doctor V64!\n", progname);
    fprintf(stderr, "Please check cables and connection.\n");
    exit(1);
  }
  receiveBytesToFile(port, file1, slen, v64format);
  fclose(file1);
  printf("\nFile transmitted successfully!\n");
  return 0;
}

int main(int argc, char *argv[])
{
  char *fname1=NULL;
  long slen=-1;
  int swapped=-1;
  unsigned int port=0;
  int mode=MODE_UNDEF;
  char romParameters[MAX_STRING_LENGTH];

  progname=argv[0];
  printf("TALK64 V1.5   Copyright (C) 1997 Andreas Sterbenz (stan@sbox.tu-graz.ac.at)\n");
  printf("This program is released under the terms of the GNU public license. NO WARRANTY\n\n");

#ifdef DEBUG_NOCOMM
  srand(0);
#endif

  {
    int i;
    char *c;
    char tmpstr[MAX_STRING_LENGTH];

    biosVersion = 0x0133;
    memset(romParameters, 0, sizeof(romParameters));
    for( i=1; i<argc; i++ ) {
      if( argv[i][0] == '-' ) {
        strncpy(tmpstr, argv[i]+1, MAX_STRING_LENGTH-1);
        tmpstr[MAX_STRING_LENGTH-1] = '\0';
        c = tmpstr;
        if( strstr(c, "sen") == c ) {
          mode = MODE_SEND;
          continue;
        }
        if( strstr(c, "rec") == c ) {
          mode = MODE_RECEIVE;
          continue;
        }
        if( strstr(c, "int") == c ) {
          mode = MODE_INTERACTIVE;
          continue;
        }
        if( strstr(c, "bios") == c ) {
          char t2[MAX_STRING_LENGTH];
          char *d=t2;
          c += strlen("bios");

          while( *c != '\0' ) {
            if( isdigit(*c) ) *d++ = *c;
            c++;
          }
          *d = '\0';
          if( sscanf(t2, "%x", &biosVersion) != 1 ) usage(progname);
          if( (biosVersion < 0x0100) || (biosVersion > 0x0999) ) {
            fprintf(stderr, "%s: Please specify the BIOS version with three digits, e.g. -bios122.\n", progname);
            exit(1);
          }
          continue;
        }
        if( *c=='o' || *c=='s' ) {
          if( *(c+1) != '\0' ) usage(progname);
          if( *c=='o' ) {
            swapped = 0;
          } else {
            swapped = 1;
          }
        } else
        if( *c=='p' ) {
          if( (strlen(c)<2) || (strlen(c)>BUFSIZE) ) usage(progname);
          if( (sscanf(c+1, "%x%s", &port, buffer1) != 1) ) usage(progname);
        } else
        if( *c=='l' ) {
          if( *(c+1) == '\0' ) {
            slen = -2;
            continue;
          }
          if( strlen(c)>BUFSIZE ) usage(progname);
          if( sscanf(c+1, "%ld%s", &slen, buffer1) != 1 ) {
            if( sscanf(c+1, "%ldM%s", &slen, buffer1) == 1 ) {
              slen <<= 17;
            } else usage(progname);
          }
        } else if( (*c == 'a') && (*(c+1) == '\0') ) {
          int j, k, l, p=0;
          for( j=i+1; j<argc; j++ ) {
            l = strlen(argv[j]);
            for( k=0; (k<l) && (p<MAX_STRING_LENGTH-1); k++, p++ ) {
              romParameters[p] = argv[j][k];
            }
            if( (j < argc-1) && (p<MAX_STRING_LENGTH-1) ) romParameters[p++] = ' ';
          }
          romParameters[HEADER_NAME_LEN-1] = '\0';
          break;
        } else usage(progname);
      } else {
        if( fname1 == NULL ) {
          fname1 = argv[i];
        } else {
          usage(progname);
        }
      }
    }
  }

  if( mode == MODE_UNDEF ) mode = MODE_SEND;
  if( mode == MODE_INTERACTIVE ) {
    if( fname1 != NULL ) usage(progname);
  } else {
    if( (mode != MODE_SEND) || (romParameters[0] == '\0') ) {
      if( fname1 == NULL ) usage(progname);
    }
  }

  port = getParPort(port);
  if( port == 0 ) {
    fflush(stdout);
    fprintf(stderr, "%s: Could not detect a parallel port, use commandline to specify one.\n", progname);
    exit(1);
  }
  printf("Using parallel port at %X; ", port);
  if( biosVersion >= 0x0130 ) {
    printf("BIOS 1.3%c extended mode.\n", (biosVersion >= 0x0133) ? '3' : '0');
  } else {
    printf("compatibility mode for pre 1.30 BIOSes.\n");
  }
  fflush(stdout);

  if( mode == MODE_INTERACTIVE ) {
    interactiveMode(port);
  } else if( mode == MODE_SEND ) {
    if( fname1 != NULL ) {
      sendFile(port, fname1, slen, swapped);
    }
    if( romParameters[0] != '\0' ) {
      if( biosVersion < 0x0130 ) {
        fflush(stdout);
        fprintf(stderr, "%s: Argument sending not available with BIOSes earlier than 1.30!\n", progname);
        exit(1);
      }
      printf("Sending arguments...");
      fflush(stdout);
      if( (sendHeaderExtSend(port, HEADER_NAME_OFFSET, HEADER_NAME_LEN) != 0) ||
          (sendBytes(port, HEADER_NAME_LEN, romParameters, 1) != 0 ) ) {
        fprintf(stderr, "\n%s: Could not send arguments, make sure you have BIOS 1.30 or later.\n", progname);
        exit(1);
      }
      printf("done.\n");
    }
  } else { /* mode == MODE_RECEIVE */
    receiveFile(port, fname1, slen, swapped);
  }
  return 0;
}
